using System;
using System.Collections.Generic;
using lpsolve55;

namespace LinearProgramming
{
    /// <summary>
    /// Possible operators for linear constraints.
    /// </summary>
    public enum OperatorType { EQ, LE, GE };

    /// <summary>
    /// Hard linear constraint, i.e. one that must be satisfied.
    /// May render a specification infeasible.
    /// </summary>
    public class Constraint
    {
        internal LinearSpec ls;
        internal double[] coeffs;
        internal Variable[] vars;
        internal OperatorType op;
        internal double rightSide;

        public int Index
        {
            get
            {
                int i = ls.Constraints.IndexOf(this);
                if (i == -1) throw new Exception("Constraint not part of ls.Constraints.");
                return i + 1;
            }
        }

        /// <summary>
        /// Coefficients of the constraint.
        /// </summary>
        public double[] Coeffs { get { return coeffs; } }

        /// <summary>
        /// Variables of the constraint.
        /// </summary>
        public Variable[] Vars { get { return vars; } }

        /// <summary>
        /// Changes the left side of the constraint, i.e. its coefficients and variables.
        /// There must be exactly one coefficient for each variable.
        /// </summary>
        /// <param name="coeffs">the new coefficients</param>
        /// <param name="vars">the new variables</param>
        public void ChangeLeftSide(double[] coeffs, Variable[] vars)
        {
            if (coeffs.Length != vars.Length)
                throw new Exception("Number of coefficients and number of variables in a constraint must be equal.");
            this.coeffs = coeffs;
            this.vars = vars;

            int[] vindexes = new int[coeffs.Length];

            int i = 0;
            foreach (Variable v in vars)
            {
                vindexes[i] = v.Index;
                i++;
            }

            if (!lpsolve.set_rowex(ls.lp, this.Index, coeffs.Length, ref coeffs[0], ref vindexes[0]))
                throw new Exception("Error in set_rowex.");
            ls.RemovePresolved();
        }

        /// <summary>
        /// The operator used for this constraint.
        /// </summary>
        public virtual OperatorType Op
        {
            get { return op; }
            set
            {
                op = value;
                if (!lpsolve.set_constr_type(ls.lp, this.Index,
                    ((op == OperatorType.EQ) ? lpsolve.lpsolve_constr_types.EQ
                    : (op == OperatorType.GE) ? lpsolve.lpsolve_constr_types.GE
                    : lpsolve.lpsolve_constr_types.LE)))
                    throw new Exception("Error in set_constr_type.");
                ls.RemovePresolved();
            }
        }

        /// <summary>
        /// The constant value that is on the right side of the operator.
        /// </summary>
        public double RightSide
        {
            get { return rightSide; }
            set
            {
                rightSide = value;
                if (!lpsolve.set_rh(ls.lp, this.Index, rightSide))
                    throw new Exception("Error in set_rh.");
                ls.RemovePresolved();
            }
        }

        public override string ToString()
        {
            String s = "";
            if (vars.Length > 0)
                s += coeffs[0].ToString() + "*" + vars[0].ToString();
            for (int i = 1; i < vars.Length; i++)
                s += " " + ((coeffs[i] >= 0) ? "+" : "") + coeffs[i].ToString() + "*" + vars[i].ToString();
            if (Op == OperatorType.EQ) s += " = ";
            else if (Op == OperatorType.GE) s += " >= ";
            else s += " <= ";
            s += rightSide.ToString();
            return s;
        }

        protected Constraint() { }

        internal Constraint(LinearSpec ls, double[] coeffs, Variable[] vars, OperatorType op, double rightSide)
        {
            if (coeffs.Length != vars.Length)
                throw new Exception("Number of coefficients and number of variables in a constraint must be equal.");

            this.ls = ls;
            this.coeffs = (double[])coeffs.Clone();
            this.vars = vars;
            this.op = op;
            this.rightSide = rightSide;

            int[] vindexes = new int[coeffs.Length];
            int i = 0;
            foreach (Variable v in vars)
            {
                vindexes[i] = v.Index;
                i++;
            }
            if (!lpsolve.add_constraintex(ls.lp, coeffs.Length, ref coeffs[0], ref vindexes[0],
                ((op == OperatorType.EQ) ? lpsolve.lpsolve_constr_types.EQ
                : (op == OperatorType.GE) ? lpsolve.lpsolve_constr_types.GE
                : lpsolve.lpsolve_constr_types.LE),
                rightSide))
                throw new Exception("Error in add_constraintex.");

            ls.Constraints.Add(this);
        }

        /// <summary>
        /// Removes the constraint from its specification.
        /// </summary>
        public void Remove()
        {
            lpsolve.del_constraint(ls.lp, this.Index);
            ls.Constraints.Remove(this);
        }
    }

    /// <summary>
    /// Soft constraint, i.e. one that does not necessarily have to be satisfied.
    /// Use this instead of hard constraints to avoid over-constrained specifications.
    /// </summary>
    public class SoftConstraint : Constraint
    {
        readonly public Variable Dpos, Dneg;
        ObjFunctionSummand dnegSummand, dposSummand;

        /// <summary>
        /// Changes the left side of the soft constraint, i.e. its coefficients and variables.
        /// There must be exactly one coefficient for each variable.
        /// </summary>
        /// <param name="coeffs">the constraint's coefficients</param>
        /// <param name="vars">the constraint's variables</param>
        public void ChangeLeftSide(double[] coeffs, Variable[] vars)
        {
            if (coeffs.Length != vars.Length)
                throw new Exception("Number of coefficients and number of variables in a constraint must be equal.");
            this.coeffs = coeffs;
            this.vars = vars;

            int[] vindexes = new int[coeffs.Length + 2];
            double[] coeffs2 = new double[coeffs.Length + 2];
            int i = 0;
            foreach (Variable v in vars)
            {
                vindexes[i] = v.Index;
                coeffs2[i] = coeffs[i];
                i++;
            }

            if (op != OperatorType.LE)
            {
                vindexes[i] = dnegSummand.Var.Index;
                coeffs2[i] = 1.0;
                i++;
            }
            if (op != OperatorType.GE)
            {
                vindexes[i] = dposSummand.Var.Index;
                coeffs2[i] = -1.0;
                i++;
            }

            if (!lpsolve.set_rowex(ls.lp, this.Index, i, ref coeffs2[0], ref vindexes[0]))
                throw new Exception("Error in set_rowex.");
            ls.RemovePresolved();
        }

        /// <summary>
        /// The operator used for this constraint.
        /// </summary>
        public override OperatorType Op
        {
            get { return op; }
            set
            {
                base.Op = value;
                ChangeLeftSide(coeffs, vars);
            }
        }

        /// <summary>
        /// The penalty coefficient for positive deviations from the soft constraint's exact solution,
        /// i.e. if the left side is too large.
        /// </summary>
        public double PenaltyNeg
        {
            get { return dnegSummand.Coeff; }
            set
            {
                if (value == dnegSummand.Coeff) return;
                dnegSummand.Coeff = value;
                ls.UpdateObjFunction();
            }
        }

        /// <summary>
        /// The penalty coefficient for negative deviations from the soft constraint's exact solution,
        /// i.e. if the left side is too small.
        /// </summary>
        public double PenaltyPos
        {
            get { return dposSummand.Coeff; }
            set
            {
                if (value == dposSummand.Coeff) return;
                dposSummand.Coeff = value;
                ls.UpdateObjFunction();
            }
        }

        public override string ToString()
        {
            return base.ToString()
                + "; PenaltyNeg=" + PenaltyNeg
                + "; PenaltyPos=" + PenaltyPos;
        }

        internal SoftConstraint(LinearSpec ls, double[] coeffs, Variable[] vars, OperatorType op, double rightSide,
            double penaltyNeg, double penaltyPos)
        {
            if (coeffs.Length != vars.Length)
                throw new Exception("Number of coefficients and number of variables in a constraint must be equal.");

            this.ls = ls;
            this.coeffs = (double[])coeffs.Clone();
            this.vars = vars;
            this.op = op;
            this.rightSide = rightSide;
            dnegSummand = ls.AddObjFunctionSummand(penaltyNeg, Dneg = new Variable(ls));
            dposSummand = ls.AddObjFunctionSummand(penaltyPos, Dpos = new Variable(ls));

            int[] vindexes = new int[coeffs.Length + 2];
            double[] coeffs2 = new double[coeffs.Length + 2];
            int i = 0;
            foreach (Variable v in vars)
            {
                vindexes[i] = v.Index;
                coeffs2[i] = coeffs[i];
                i++;
            }
            if (op != OperatorType.LE)
            {
                vindexes[i] = dnegSummand.Var.Index;
                coeffs2[i] = 1.0;
                i++;
            }
            if (op != OperatorType.GE)
            {
                vindexes[i] = dposSummand.Var.Index;
                coeffs2[i] = -1.0;
                i++;
            }

            if (!lpsolve.add_constraintex(ls.lp, i, ref coeffs2[0], ref vindexes[0],
                ((op == OperatorType.EQ) ? lpsolve.lpsolve_constr_types.EQ
                : (op == OperatorType.GE) ? lpsolve.lpsolve_constr_types.GE
                : lpsolve.lpsolve_constr_types.LE),
                rightSide))
                throw new Exception("Error in add_constraintex.");

            ls.Constraints.Add(this);
        }

        /// <summary>
        /// Removes the soft constraint from its specification.
        /// </summary>
        public void Remove()
        {
            dnegSummand.Var.Remove();
            dposSummand.Var.Remove();
            dnegSummand.Remove();
            dposSummand.Remove();
            base.Remove();
        }
    }
}
