using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using lpsolve55;

namespace LinearProgramming
{
    /// <summary>
    /// A variable in a linear specification.
    /// </summary>
    public class Variable
    {
        public LinearSpec ls;
        double min = 0;
        double max = Double.MaxValue;

        public int Index
        {
            get
            {
                int i = ls.Variables.IndexOf(this);
                if (i == -1) throw new Exception("Variable not part of ls.Variables.");
                return i + 1;
            }
        }
        public double Value = Double.NaN;

        /// <summary>
        /// Minimum value of the variable.
        /// </summary>
        public double Min
        {
            get { return min; }
            set
            {
                min = value;
                lpsolve.set_bounds(ls.lp, this.Index, min, max);
            }
        }

        /// <summary>
        /// Maximum value of the variable.
        /// </summary>
        public double Max
        {
            get { return max; }
            set
            {
                max = value;
                lpsolve.set_bounds(ls.lp, this.Index, min, max);
            }
        }

        /// <summary>
        /// Sets the minimum and maximum values of the variable.
        /// </summary>
        /// <param name="min">minimum value</param>
        /// <param name="max">maximum value</param>
        public void SetRange(double min, double max)
        {
            this.min = min;
            this.max = max;
            lpsolve.set_bounds(ls.lp, this.Index, min, max);
        }

        public override string ToString()
        {
            return "Var"+Index;
        }

        /// <summary>
        /// Adds a constraint that sets this variable equal to the given one.
        /// </summary>
        /// <param name="v">variable that should have the same value</param>
        /// <returns>the new equality constraint</returns>
        public Constraint IsEqual(Variable v)
        {
            return ls.AddConstraint(new double[] { 1, -1 }, new Variable[] { this, v }, OperatorType.EQ, 0);
        }

        /// <summary>
        /// Adds a constraint that sets this variable smaller or equal to the given one.
        /// </summary>
        /// <param name="v">variable that should have a larger or equal value</param>
        /// <returns>the new constraint</returns>
        public Constraint IsSmallerOrEqual(Variable v)
        {
            return ls.AddConstraint(new double[] { 1, -1 }, new Variable[] { this, v }, OperatorType.LE, 0);
        }

        /// <summary>
        /// Adds a constraint that sets this variable greater or equal to the given one.
        /// </summary>
        /// <param name="v">variable that should have a smaller or equal value</param>
        /// <returns>the new constraint</returns>
        public Constraint IsGreaterOrEqual(Variable v)
        {
            return ls.AddConstraint(new double[] { -1, 1 }, new Variable[] { v, this }, OperatorType.GE, 0);
        }

        internal protected Variable(LinearSpec ls)
        {
            this.ls = ls;
            ls.Variables.Add(this);
            if (ls.Variables.Count > ls.columns)
            {
                double d = 0;
                int i = 0;
                if (!lpsolve.add_columnex(ls.lp, 0, ref d, ref i))
                    throw new Exception("Error in add_columnex.");
            }
        }

        /// <summary>
        /// Removes the variable from its specification.
        /// </summary>
        public void Remove()
        {
            lpsolve.del_column(ls.lp, this.Index);
            ls.Variables.Remove(this);
        }
    }
    
    /// <summary>
    /// The two possibilities for optimizing the objective function.
    /// </summary>
    public enum OptimizationType { MINIMIZE, MAXIMIZE };

    /// <summary>
    /// The possible results of a solving attempt.
    /// </summary>
    public enum ResultType
    {
        NOMEMORY = -2, ERROR = -1, OPTIMAL = 0, SUBOPTIMAL = 1, INFEASIBLE = 2, UNBOUNDED = 3,
        DEGENERATE = 4, NUMFAILURE = 5, USERABORT = 6, TIMEOUT = 7, PRESOLVED = 9, PROCFAIL = 10,
        PROCBREAK = 11, FEASFOUND = 12, NOFEASFOUND = 13
    };

    /// <summary>
    /// Specification of a linear programming problem.
    /// </summary>
    public class LinearSpec
    {   
        public int lp;
        int lp_presolved = -1;
        internal int columns = 0;
        OptimizationType optimization = OptimizationType.MINIMIZE;

        /// <summary>
        /// The summands of the objective function.
        /// </summary>
        public List<ObjFunctionSummand> ObjFunctionSummands = new List<ObjFunctionSummand>();
        
        /// <summary>
        /// The variables of the specification.
        /// </summary>
        public List<Variable> Variables = new List<Variable>();

        /// <summary>
        /// The constraints of the specification.
        /// </summary>
        public List<Constraint> Constraints = new List<Constraint>();

        /// <summary>
        /// The result of the last solving attempt.
        /// </summary>
        public ResultType Result = ResultType.ERROR;

        /// <summary>
        /// The value of the objective function at the end of the last solving attempt.
        /// </summary>
        public double ObjectiveValue = Double.NaN;

        /// <summary>
        /// The duration of the last solving attempt in milliseconds.
        /// </summary>
        public double SolvingTime = Double.NaN;

        /// <summary>
        /// Creates a new specification for a linear programming problem.
        /// </summary>
        public LinearSpec()
        {
            lp = lpsolve.make_lp(0, 0);
            if (lp == 0) throw new Exception("Couldn't construct a new model.");
            lpsolve.set_verbose(lp, 1); // only print critical messages

            // scaling and anti_degen did not seem to help with NUMFAILURE
            //lpsolve.set_scaling(lp, lpsolve.lpsolve_scales.SCALE_CURTISREID | lpsolve.lpsolve_scales.SCALE_DYNUPDATE);
            //lpsolve.set_anti_degen(lp,
            //    lpsolve.lpsolve_anti_degen.ANTIDEGEN_INFEASIBLE
            //    | lpsolve.lpsolve_anti_degen.ANTIDEGEN_STALLING
            //    | lpsolve.lpsolve_anti_degen.ANTIDEGEN_INFEASIBLE
            //    | lpsolve.lpsolve_anti_degen.ANTIDEGEN_NUMFAILURE);
        }

        public void UpdateObjFunction()
        {
            double[] coeffs = new double[ObjFunctionSummands.Count];
            int[] vindexes = new int[ObjFunctionSummands.Count];
            int i = 0;
            foreach (ObjFunctionSummand s in ObjFunctionSummands) {
                coeffs[i] = s.Coeff;
                vindexes[i] = s.Var.Index;
                i++;
            }
            if (!lpsolve.set_obj_fnex(lp, ObjFunctionSummands.Count, ref coeffs[0], ref vindexes[0]))
                throw new Exception("Error in set_obj_fnex.");

            RemovePresolved();
        }

        /// <summary>
        /// Sets the objective function.
        /// </summary>
        /// <param name="coeffs">the objective function's coefficients</param>
        /// <param name="vars">the objective function's variables</param>
        public void SetObjFunction(double[] coeffs, Variable[] vars)
        {
            if (coeffs.Length != vars.Length)
                throw new Exception("Number of coefficients and number of variables in objective function must be equal.");
            
            ObjFunctionSummands = new List<ObjFunctionSummand>();
            for(int i=0; i<coeffs.Length; i++)
                ObjFunctionSummands.Add(new ObjFunctionSummand(this, coeffs[i], vars[i]));

            UpdateObjFunction();
        }

        /// <summary>
        /// Adds a new summand to the objective function.
        /// </summary>
        /// <param name="coeff">the summand's coefficient</param>
        /// <param name="var">the summand's variable</param>
        /// <returns>the new summand</returns>
        public ObjFunctionSummand AddObjFunctionSummand(double coeff, Variable var)
        {
            ObjFunctionSummand s = new ObjFunctionSummand(this, coeff, var);
            ObjFunctionSummands.Add(s);
            UpdateObjFunction();
            return s;
        }

        /// <summary>
        /// Sets wether the solver should minimize or maximize the objectibe function.
        /// The default is mnimization.
        /// </summary>
        public OptimizationType Optimization
        {
            get { return optimization; }
            set
            {
                if (optimization == OptimizationType.MINIMIZE)
                    lpsolve.set_minim(lp);
                else
                    lpsolve.set_maxim(lp);

                RemovePresolved();
            }
        }

        /// <summary>
        /// Adds a new variable to the specification.
        /// </summary>
        /// <returns>the new variable</returns>
        public Variable AddVariable()
        {
            return new Variable(this);
        }

        /// <summary>
        /// Adds a new hard linear constraint to the specification.
        /// </summary>
        /// <param name="coeffs">the constraint's coefficients</param>
        /// <param name="vars">the constraint's variables</param>
        /// <param name="op">the constraint's operand</param>
        /// <param name="rightSide">the constant value on the constraint's right side</param>
        /// <returns>the new constraint</returns>
        public Constraint AddConstraint(double[] coeffs, Variable[] vars, OperatorType op, double rightSide)
        {
            Constraint c = new Constraint(this, coeffs, vars, op, rightSide);
            RemovePresolved();
            return c;
        }

        /// <summary>
        /// Adds a new soft linear constraint to the specification,
        /// i.e. a constraint that does not always have to be satisfied.
        /// </summary>
        /// <param name="coeffs">the constraint's coefficients</param>
        /// <param name="vars">the constraint's variables</param>
        /// <param name="op">the constraint's operand</param>
        /// <param name="rightSide">the constant value on the constraint's right side</param>
        /// <param name="penaltyNeg">the coefficient penalizing negative deviations from the exact solution</param>
        /// <param name="penaltyPos">the coefficient penalizing positive deviations from the exact solution</param>
        /// <returns></returns>
        public SoftConstraint AddSoftConstraint(double[] coeffs, Variable[] vars, OperatorType op, double rightSide,
            double penaltyNeg, double penaltyPos)
        {
            SoftConstraint c = new SoftConstraint(this, coeffs, vars, op, rightSide, penaltyNeg, penaltyPos);
            RemovePresolved();
            return c;
        }

        public PenaltyFunction AddPenaltyFunction(Variable var, double[] xs, double[] gs)
        {
            return new PenaltyFunction(this, var, xs, gs);
        }

        /// <summary>
        /// Remove a cached presolved model, if existent.
        /// This is automatically done each time after the model has been changed, 
        /// to avoid an old cached presolved model getting out of sync.
        /// </summary>
        public void RemovePresolved()
        {
            if (lp_presolved == -1) return;
            lpsolve.delete_lp(lp_presolved);
            lp_presolved = -1;
        }

        /// <summary>
        /// Creates and caches a simplified version of the linear programming problem, 
        /// where redundant rows, columns and constraints are removed, 
        /// if it has not been created before.
        /// Then, the simplified problem is solved.
        /// </summary>
        /// <returns>the result of the solving attempt</returns>
        public ResultType Presolve()
        {
            long start, end, freq;
            PrecisionTimer.QueryPerformanceCounter(out start);

            if (lp_presolved == -1)
            {
                lp_presolved = lpsolve.copy_lp(lp);
                lpsolve.set_presolve(lp_presolved,
                        lpsolve.lpsolve_presolve.PRESOLVE_ROWS
                        | lpsolve.lpsolve_presolve.PRESOLVE_COLS
                        | lpsolve.lpsolve_presolve.PRESOLVE_LINDEP,
                        lpsolve.get_presolveloops(lp_presolved));
            }

            Result = (ResultType)lpsolve.solve(lp_presolved);
            ObjectiveValue = lpsolve.get_objective(lp_presolved);

            if (Result == ResultType.OPTIMAL)
            {
                foreach (Variable v in Variables)
                    v.Value = lpsolve.get_var_primalresult(lp_presolved,
                        lpsolve.get_Norig_rows(lp_presolved) + v.Index);
            }

            PrecisionTimer.QueryPerformanceCounter(out end);
            PrecisionTimer.QueryPerformanceFrequency(out freq);
            SolvingTime = ((double)(end - start) / (double)freq) * 1000.0;
            return Result;
        }

        /// <summary>
        /// Tries to solve the linear programming problem.
        /// If a cached simplified version of the problem exists, it is used instead.
        /// </summary>
        /// <returns>the result of the solving attempt</returns>
        public ResultType Solve()
        {
            if (lp_presolved != -1) return Presolve();

            long start, end, freq;
            PrecisionTimer.QueryPerformanceCounter(out start);

            Result = (ResultType)lpsolve.solve(lp);
            ObjectiveValue = lpsolve.get_objective(lp);

            if (Result == ResultType.OPTIMAL)
            {
                double[] x = new double[Variables.Count];
                if (!lpsolve.get_variables(lp, ref x[0]))
                    throw new Exception("Error in get_variables.");
                int i = 0;
                foreach (Variable v in Variables) v.Value = x[i++];
            }
            PrecisionTimer.QueryPerformanceCounter(out end);
            PrecisionTimer.QueryPerformanceFrequency(out freq);
            SolvingTime = ((double)(end - start) / (double)freq) * 1000.0;
            return Result;
        }

        /// <summary>
        /// Writes the specification into a text file.
        /// The file will be overwritten if it exists.
        /// </summary>
        /// <param name="fname">the file name</param>
        public void Save(String fname){
            lpsolve.write_lp(lp, fname);
        }

        /// <summary>
        /// Removes the specification.
        /// </summary>
        public void Remove()
        {
            lpsolve.delete_lp(lp);
        }
    }
}

/// <summary>
/// Class for accessing precise timing information.
/// </summary>
public class PrecisionTimer
{
    [DllImport("Kernel32.dll")]
    public static extern bool QueryPerformanceCounter(out long value);

    [DllImport("Kernel32.dll")]
    public static extern bool QueryPerformanceFrequency(out long value);
}
